

# Standard library imports.
import logging

# Enthought library imports.
from envisage.ui.workbench.api import WorkbenchApplication
from envisage.core_plugin import CorePlugin
from envisage.ui.workbench.workbench_plugin import WorkbenchPlugin
from enthought.epdlab.plugins.code_editor.plugin import CodeEditorPlugin

# Logging.
logger = logging.getLogger()
logger.addHandler(logging.StreamHandler(open('update_checker.log', 'w')))
logger.setLevel(logging.DEBUG)


def main():
    application = Application(
        id='Update Checker Tester', plugins=[UpdateChecker()]
    )

    application = WorkbenchApplication(
        id='update_checker_tester', name='Update Checker',
        plugins=[
            CorePlugin(),
            WorkbenchPlugin(),
            UpdateCheckerPlugin(

                ),
        ])
    application.run()

if __name__ == "__main__":
    main()



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2006 by Enthought, Inc.
#  All rights reserved.
#
#  Author: Dave Peterson <dpeterson@enthought.com>
#
#-----------------------------------------------------------------------------

# IDs of services provided by this plugin
from .services import IPROJECT_MODEL, IPROJECT_UI

# Commonly referred to classes within this plugin
from .factory_definition import FactoryDefinition
from .model_service import ModelService
from .project import Project
from .project_action import ProjectAction
from .project_factory import ProjectFactory
from .view.project_view import ProjectView
# FIXME: Add back this import when it actually works :)
#from .editor.project_editor import ProjectEditor

#### EOF #####################################################################



""" An implementation of a client controlling a remote editor, but also using
    envisage to retrieve an execution engine, and run the commands from the
    editor.
"""

# Standard library imports
import logging

# Enthought library imports
from traits.api import Instance
from envisage.api import Application
from envisage.plugins.python_shell.api import IPythonShell
from pyface.api import GUI

# Local imports
from envisage.plugins.remote_editor.remote_editor_controller import \
    RemoteEditorController

logger = logging.getLogger(__name__)


class EnvisageRemoteEditorController(RemoteEditorController):
    """ An implementation of a client controlling a remote editor, but also
        using envisage to retrieve an execution engine, and run the commands
        from the editor.
    """

    # Tell the Client code to play well with the wx or Qt4 event loop
    ui_dispatch = 'auto'

    # A reference to the Envisage application.
    application = Instance(Application)

    ###########################################################################
    # EnvisageRemoteEditorController interface
    ###########################################################################

    def run_file(self, path):
        """ Called by the server to execute a file.
        """
        shell = self.application.get_service(IPythonShell)
        shell.execute_file(path, hidden=False)

    def run_text(self, text):
        """ Called by the server to execute text.
        """
        shell = self.application.get_service(IPythonShell)
        shell.execute_command(text, hidden=False)

    ###########################################################################
    # Enshell Client interface
    ###########################################################################

    def handle_command(self, command, arguments):
        """ Hande commands coming in from the server.
        """
        if command == "run_file":
            self.run_file(arguments)
            return True
        elif command == "run_text":
            self.run_text(arguments)
            return True
        return False




""" Tests for plugins. """


# Standard library imports.
from os.path import exists, join

# Enthought library imports.
from envisage.api import Application, ExtensionPoint
from envisage.api import IPluginActivator, Plugin, contributes_to
from traits.api import HasTraits, Instance, Int, Interface, List
from traits.api import provides
from traits.testing.unittest_tools import unittest


def listener(obj, trait_name, old, new):
    """ A useful trait change handler for testing! """

    listener.obj = obj
    listener.trait_name = trait_name
    listener.old = old
    listener.new = new

    return


class TestApplication(Application):
    """ The type of application used in the tests. """

    id = 'test'


class PluginTestCase(unittest.TestCase):
    """ Tests for plugins. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        return

    def tearDown(self):
        """ Called immediately after each test method has been called. """

        return

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_id_policy(self):
        """ id policy """

        # If no Id is specified then use 'module_name.class_name'.
        p = Plugin()
        self.assertEqual('envisage.plugin.Plugin', p.id)

        # If an Id is specified make sure we use it!
        p = Plugin(id='wilma')
        self.assertEqual('wilma', p.id)

        # Make sure setting the name doesn't interfere with the Id.
        p = Plugin(name='fred', id='wilma')
        self.assertEqual('wilma', p.id)
        self.assertEqual('fred', p.name)

        return

    def test_name_policy(self):
        """ name policy """

        # If name is specified then use the plugin's class name.
        p = Plugin()
        self.assertEqual('Plugin', p.name)

        # If a name is specified make sure we use it!
        p = Plugin(name='wilma')
        self.assertEqual('wilma', p.name)

        # Try a camel case plugin class.
        class ThisIsMyPlugin(Plugin):
            pass

        p = ThisIsMyPlugin()
        self.assertEqual('This Is My Plugin', p.name)

        return

    def test_plugin_activator(self):
        """ plugin activator. """

        @provides(IPluginActivator)
        class NullPluginActivator(HasTraits):
            """ A plugin activator that does nothing! """

            def start_plugin(self, plugin):
                """ Start a plugin. """

                self.started = plugin

                return

            def stop_plugin(self, plugin):
                """ Stop a plugin. """

                self.stopped = plugin

                return

        class PluginA(Plugin):
            id = 'A'

        class PluginB(Plugin):
            id = 'B'

        plugin_activator = NullPluginActivator()

        a = PluginA(activator=plugin_activator)
        b = PluginB()

        application = TestApplication(plugins=[a, b])
        application.start()

        # Make sure A's plugin activator was called.
        self.assertEqual(a, plugin_activator.started)

        # Stop the application.
        application.stop()

        # Make sure A's plugin activator was called.
        self.assertEqual(a, plugin_activator.stopped)

        return

    def test_service(self):
        """ service """

        class Foo(HasTraits):
            pass

        class Bar(HasTraits):
            pass

        class Baz(HasTraits):
            pass

        class PluginA(Plugin):
            id = 'A'
            foo = Instance(Foo, (), service=True)
            bar = Instance(Bar, (), service=True)
            baz = Instance(Baz, (), service=True)

        a = PluginA()

        application = TestApplication(plugins=[a])
        application.start()

        # Make sure the services were registered.
        self.assertNotEqual(None, application.get_service(Foo))
        self.assertEqual(a.foo, application.get_service(Foo))

        self.assertNotEqual(None, application.get_service(Bar))
        self.assertEqual(a.bar, application.get_service(Bar))

        self.assertNotEqual(None, application.get_service(Baz))
        self.assertEqual(a.baz, application.get_service(Baz))

        application.stop()

        # Make sure the service was unregistered.
        self.assertEqual(None, application.get_service(Foo))
        self.assertEqual(None, application.get_service(Bar))
        self.assertEqual(None, application.get_service(Baz))

        return

    def test_service_protocol(self):
        """ service protocol """

        class IFoo(Interface):
            pass

        class IBar(Interface):
            pass

        @provides(IFoo, IBar)
        class Foo(HasTraits):
            pass

        class PluginA(Plugin):
            id = 'A'
            foo = Instance(Foo, (), service=True, service_protocol=IBar)

        a = PluginA()

        application = TestApplication(plugins=[a])
        application.start()

        # Make sure the service was registered with the 'IBar' protocol.
        self.assertNotEqual(None, application.get_service(IBar))
        self.assertEqual(a.foo, application.get_service(IBar))

        application.stop()

        # Make sure the service was unregistered.
        self.assertEqual(None, application.get_service(IBar))

        return

    def test_multiple_trait_contributions(self):
        """ multiple trait contributions """

        class PluginA(Plugin):
            id = 'A'
            x  = ExtensionPoint(List, id='x')

        class PluginB(Plugin):
            id = 'B'

            x  = List([1, 2, 3], contributes_to='x')
            y  = List([4, 5, 6], contributes_to='x')

        a = PluginA()
        b = PluginB()

        application = TestApplication(plugins=[a, b])

        # We should get an error because the plugin has multiple traits
        # contributing to the same extension point.
        self.failUnlessRaises(ValueError, application.get_extensions, 'x')

        return

    def test_exception_in_trait_contribution(self):
        """ exception in trait contribution """

        class PluginA(Plugin):
            id = 'A'
            x  = ExtensionPoint(List, id='x')

        class PluginB(Plugin):
            id = 'B'

            x  = List(contributes_to='x')

            def _x_default(self):
                """ Trait initializer. """

                raise 1/0

        a = PluginA()
        b = PluginB()

        application = TestApplication(plugins=[a, b])

        # We should get an when we try to get the contributions to the
        # extension point.
        self.failUnlessRaises(
            ZeroDivisionError, application.get_extensions, 'x'
        )

        return

    def test_contributes_to(self):
        """ contributes to """

        class PluginA(Plugin):
            id = 'A'
            x  = ExtensionPoint(List, id='x')

        class PluginB(Plugin):
            id = 'B'
            x  = List([1, 2, 3], contributes_to='x')

        a = PluginA()
        b = PluginB()

        application = TestApplication(plugins=[a, b])

        # We should get an error because the plugin has multiple traits
        # contributing to the same extension point.
        self.assertEqual([1, 2, 3], application.get_extensions('x'))

        return

    def test_contributes_to_decorator(self):
        """ contributes to decorator """

        class PluginA(Plugin):
            id = 'A'
            x  = ExtensionPoint(List, id='x')

        class PluginB(Plugin):
            id = 'B'

            @contributes_to('x')
            def _x_contributions(self):
                return [1, 2, 3]

        a = PluginA()
        b = PluginB()

        application = TestApplication(plugins=[a, b])
        self.assertEqual([1, 2, 3], application.get_extensions('x'))

        return

    def test_contributes_to_decorator_ignored_if_trait_present(self):
        """ contributes to decorator ignored if trait present """

        class PluginA(Plugin):
            id = 'A'
            x  = ExtensionPoint(List, id='x')

        class PluginB(Plugin):
            id = 'B'
            x  = List([1, 2, 3], contributes_to='x')

            @contributes_to('x')
            def _x_contributions(self):
                return [4, 5, 6]

        a = PluginA()
        b = PluginB()

        application = TestApplication(plugins=[a, b])
        self.assertEqual([1, 2, 3], application.get_extensions('x'))

        return

    def test_add_plugins_to_empty_application(self):
        """ add plugins to empty application """

        class PluginA(Plugin):
            id = 'A'
            x  = ExtensionPoint(List(Int), id='x')

            def _x_items_changed(self, event):
                self.added   = event.added
                self.removed = event.removed

                return

        class PluginB(Plugin):
            id = 'B'
            x  = List(Int, [1, 2, 3], contributes_to='x')

        class PluginC(Plugin):
            id = 'C'
            x  = List(Int, [4, 5, 6], contributes_to='x')

        a = PluginA()
        b = PluginB()
        c = PluginC()

        # Create an empty application.
        application = TestApplication()
        application.start()

        # Add the plugin that offers the extension point.
        application.add_plugin(a)

        #######################################################################
        # fixme: Currently, we connect up extension point traits when the
        # plugin is started. Is this right? Should we start plugins by default
        # when we add them (and maybe have the ability to add a plugin without
        # starting it?).
        #
        # I think we should start the plugin, otherwise you have the wierdness
        # that the extension contributed by the plugin are available after
        # the call to 'add_plugin', but the plugin isn't started?!?
        #######################################################################

        application.start_plugin(a)

        #######################################################################
        # fixme: Currently, we only fire changed events if an extension point
        # has already been accessed! Is this right?
        #######################################################################

        self.assertEqual([], a.x)

        # Add a plugin that contributes to the extension point.
        application.add_plugin(b)

        # Make sure that we pick up B's extensions and that the appropriate
        # trait event was fired.
        self.assertEqual([1, 2, 3], a.x)
        self.assertEqual([1, 2, 3], a.added)

        # Add another plugin that contributes to the extension point.
        application.add_plugin(c)

        self.assertEqual([1, 2, 3, 4, 5, 6], a.x)
        self.assertEqual([4, 5, 6], a.added)

        # Remove the first contributing plugin.
        application.remove_plugin(b)

        self.assertEqual([4, 5, 6], a.x)
        self.assertEqual([1, 2, 3], a.removed)

        # Remove the second contributing plugin.
        application.remove_plugin(c)

        self.assertEqual([], a.x)
        self.assertEqual([4, 5, 6], a.removed)

        return

    def test_home(self):
        """ home """

        class PluginA(Plugin):
            id = 'A'

        class PluginB(Plugin):
            id = 'B'

        a = PluginA()
        b = PluginB()

        application = TestApplication(plugins=[a, b])

        # Make sure that each plugin gets its own directory.
        self.assertEqual(join(application.home, 'plugins', a.id), a.home)
        self.assertEqual(join(application.home, 'plugins', b.id), b.home)

        # Make sure that the directories got created.
        self.assert_(exists(a.home))
        self.assert_(exists(b.home))

        # Create a new application with plugins with the same Id to make sure
        # that it all works when the directories already exist.
        a = PluginA()
        b = PluginB()

        application = TestApplication(plugins=[a, b])

        # Make sure that each plugin gets its own directory.
        self.assertEqual(join(application.home, 'plugins', a.id), a.home)
        self.assertEqual(join(application.home, 'plugins', b.id), b.home)

        # Make sure the directories got created.
        self.assert_(exists(a.home))
        self.assert_(exists(b.home))

        return


# Entry point for stand-alone testing.
if __name__ == '__main__':
    unittest.main()

#### EOF ######################################################################



""" The Developer perspective. """


# Enthought library imports.
from pyface.workbench.api import Perspective, PerspectiveItem


class DeveloperPerspective(Perspective):
    """ The Developer perspective.

    This perspective is intented to contain views and editors useful for
    inspecting and debugging a running Envisage application.

    """

    # The root of all view Ids in this package.
    ROOT = 'envisage.developer.ui.view'

    # View Ids.
    APPLICATION_BROWSER_VIEW = ROOT + '.application_browser_view'
    EXTENSION_REGISTRY_BROWSER_VIEW = ROOT + '.extension_registry_browser_view'
    SERVICE_REGISTRY_BROWSER_VIEW = ROOT + '.service_registry_browser_view'

    # The perspective's name.
    name = 'Developer'

    # Should the editor area be shown in this perspective?
    show_editor_area = True

    # The contents of the perspective.
    contents = [
        PerspectiveItem(
            id       = APPLICATION_BROWSER_VIEW,
            position = 'left'
        ),

        PerspectiveItem(
            id          = EXTENSION_REGISTRY_BROWSER_VIEW,
            position    = 'bottom',
            relative_to = APPLICATION_BROWSER_VIEW
        ),

        PerspectiveItem(
            id          = 'Python',
            position    = 'bottom',
        ),

        PerspectiveItem(
            id          = SERVICE_REGISTRY_BROWSER_VIEW,
            position    = 'right',
        ),
    ]

#### EOF ######################################################################



""" A view containing the contents of a Python shell namespace. """

from __future__ import print_function

# Enthought library imports.
from envisage.plugins.python_shell.api import IPythonShell
from envisage.plugins.ipython_shell.api import INamespaceView
from pyface.workbench.api import View
from traits.api import Property, provides, Instance, \
    Str, HasTraits
from traitsui.api import Item, TreeEditor, Group
from traitsui.api import View as TraitsView
from traitsui.value_tree import DictNode, StringNode, \
        value_tree_nodes
from pyface.timer.api import Timer
from pyface.api import GUI

def search_namespace(namespace, string, depth=3):
    """ Iterator on a dictionnary-like object.

        Given a namespace, search recursively for a name containing the
        string in the enclosed modules and classes.
    """
    if depth==0:
        raise StopIteration
    for child_name in namespace:
        child = namespace[child_name]
        if string in child_name:
            yield child_name, child
        if hasattr(child, '__dict__'):
            for suitable_child_name, suitable_child in \
                    search_namespace(child.__dict__,
                                                string, depth=depth-1):
                yield ('%s.%s' % (child_name, suitable_child_name),
                                    suitable_child)


def filter_namespace(namespace, string, depth=3):
    """ Return a flattened dictionnary to the depth given, with
        only the keys matching the given name.
    """
    out_dict = dict()
    for key, item in search_namespace(namespace, string, depth=depth):
        out_dict[key] = item
    return out_dict


def explore(node):
    """ Small helper function to graphically edit an object.
    """
    # FIXME: This is really very dumb. The view used should be made
    # better. The logics should probably be put in the nodes themselves,
    # subclassing them.
    name = node.name
    obj  = node.value
    class MyClass(HasTraits):
        the_object = obj

        view = TraitsView(Item('the_object', style='custom', show_label=False),
                        resizable=True,
                        title=name,
                        width=600,
                        )

    return MyClass().edit_traits()

class NamespaceNode(DictNode):
    """ Subclass of the DictNode for the namespace purposes.
    """

    def tno_get_icon ( self, node, is_expanded ):
        """ Returns the icon for a specified object.

            We overwrite this method because we don't have a default icon for
            this object.
        """
        return ('@icons:dict_node')

    def tno_get_children ( self, node ):
        """ Gets the object's children.

            We overwrite this method for a nicer label on the objects.
        """
        node_for = self.node_for
        items    = list(self.value.items())
        items.sort( lambda l, r: cmp( l[0], r[0] ) )
        if len( items ) > 500:
            return ([ self.node_for( k, v ) for k, v in items[: 250 ] ] +
                    [ StringNode( value = '...', readonly = True ) ]        +
                    [ self.node_for( k, v ) for k, v in items[ -250: ] ])

        return [ self.node_for( k, v ) for k, v in items ]



################################################################################
@provides(INamespaceView)
class NamespaceView(View):
    """ A view containing the contents of the Python shell namespace. """

    #### 'IView' interface ####################################################

    # The part's globally unique identifier.
    id = 'envisage.plugins.ipython_shell.namespace_view'

    # The view's name.
    name = 'Namespace'

    # The default position of the view relative to the item specified in the
    # 'relative_to' trait.
    position = 'left'

    #### 'NamespaceView' interface ############################################

    # The different tree nodes
    tree_nodes = Property(depends_on='search_text')

    # Search text
    search_text = Str

    tree_editor = Property(depends_on="ui")

    # The timer used to refresh the ui
    _refresh_tree_nodes_timer = Instance(Timer)

    def __refresh_tree_nodes_timer_default(self):
        return Timer(100, self._refresh_tree_nodes)

    ###########################################################################
    # 'View' interface.
    ###########################################################################

    traits_view = TraitsView(
            Group(Item('search_text', label='Search')),
            Item(
                'tree_nodes',
                id     = 'table',
                editor = TreeEditor(
                                auto_open=1,
                                hide_root=True,
                                editable=False,
                                nodes=value_tree_nodes,
                                on_dclick='object._explore',
                                ),
                springy = True,
                resizable = True,
                show_label = False
            ),
            resizable = True,
        )


    def create_control(self, parent):
        """ Creates the toolkit-specific control that represents the view.

        'parent' is the toolkit-specific control that is the view's parent.

        """
        self.ui = self.edit_traits(parent=parent, kind='subpanel')

        # Register the view as a service.
        self.window.application.register_service(INamespaceView, self)

        shell = self.window.application.get_service(IPythonShell)
        if shell is not None:
            shell.on_trait_change(self._on_names_changed, 'names')
            self._on_names_changed(shell.names)

        return self.ui.control


    def destroy_control(self):
        """ Destroys the toolkit-specific control that represents the view.

        """

        super(NamespaceView, self).destroy_control()

        # Remove the namespace change handler
        shell= self.window.application.get_service(IPythonShell)
        if shell is not None:
            shell.on_trait_change(
                self._on_names_changed, 'names', remove=True
            )


    ###########################################################################
    # 'NamespaceView' interface.
    ###########################################################################

    #### Properties ###########################################################

    def _get_tree_nodes(self):
        """ Property getter. """

        shell = self.window.application.get_service(IPythonShell)

        # Cater for an un-initialized python shell view
        if shell is None:
            return NamespaceNode(value={}, readonly=True)
        filtered_namespace = dict()

        for name in shell.names:
            filtered_namespace[name] = shell.lookup(name)

        if not self.search_text == '':
            filtered_namespace = filter_namespace(filtered_namespace,
                                                        self.search_text)

        return NamespaceNode(value=filtered_namespace, readonly=True)


    def _get_tree_editor(self):
        """ Walk the editor list to retrieve the instance of the
            tree editor currently used.
        """
        for editor in self.ui._editors:
            print(editor)
        return self.ui._editors[-1]


    def _refresh_tree_nodes(self):
        """ Callback called by a timer to refresh the UI.

            The UI is refreshed by a timer to buffer the refreshes,
            in order not to slow down the execution engine.
        """
        self.trait_property_changed('tree_nodes', None)
        self._refresh_tree_nodes_timer.Stop()

    ###########################################################################
    # Private interface.
    ###########################################################################

    #### Trait change handlers ################################################

    def _on_names_changed(self, new):
        """ Dynamic trait change handler. """

        if not self._refresh_tree_nodes_timer.IsRunning():
            GUI.invoke_later(self._refresh_tree_nodes_timer.Start)

    def _explore(self, object):
        """ Displays a view of the object.
        """
        explore(object)

#### EOF ######################################################################



""" A simple plugin manager implementation. """


from fnmatch import fnmatch
import logging

from traits.api import Event, HasTraits, Instance, List, Str, provides

from .i_application import IApplication
from .i_plugin import IPlugin
from .i_plugin_manager import IPluginManager
from .plugin_event import PluginEvent



logger = logging.getLogger(__name__)


@provides(IPluginManager)
class PluginManager(HasTraits):
    """ A simple plugin manager implementation.

    This implementation manages an explicit collection of plugin instances,
    e.g::

        plugin_manager = PluginManager(
             plugins = [
                 MyPlugin(),
                 YourPlugin()
             ]
        )

    Plugins can be added and removed after construction time via the methods
    'add_plugin' and 'remove_plugin'.

    """

    #### 'IPluginManager' protocol #############################################

    # Fired when a plugin has been added to the manager.
    plugin_added = Event(PluginEvent)

    # Fired when a plugin has been removed from the manager.
    plugin_removed = Event(PluginEvent)

    #### 'PluginManager' protocol ##############################################

    # The application that the plugin manager is part of.
    application = Instance(IApplication)
    def _application_changed(self, trait_name, old, new):
        """ Static trait change handler. """

        self._update_plugin_application([], self._plugins)

        return

    # An optional list of the Ids of the plugins that are to be excluded by
    # the manager.
    #
    # Each item in the list is actually an 'fnmatch' expression.
    exclude = List(Str)

    # An optional list of the Ids of the plugins that are to be included by
    # the manager (i.e. *only* plugins with Ids in this list will be added to
    # the manager).
    #
    # Each item in the list is actually an 'fnmatch' expression.
    include = List(Str)

    #### 'object' protocol #####################################################

    def __init__(self, plugins=None, **traits):
        """ Constructor.

        We allow the caller to specify an initial list of plugins, but the
        list itself is not part of the public API. To add and remove plugins
        after construction, use the 'add_plugin' and 'remove_plugin' methods
        respectively. The manager is also iterable, so to iterate over the
        plugins use 'for plugin in plugin_manager'.

        """

        super(PluginManager, self).__init__(**traits)

        if plugins is not None:
            self._plugins = plugins

        return

    def __iter__(self):
        """ Return an iterator over the manager's plugins. """

        plugins = [
            plugin for plugin in self._plugins

            if self._include_plugin(plugin.id)
        ]

        return iter(plugins)

    #### 'IPluginManager' protocol #############################################

    def add_plugin(self, plugin):
        """ Add a plugin to the manager. """

        self._plugins.append(plugin)
        self.plugin_added = PluginEvent(plugin=plugin)

        return

    def get_plugin(self, plugin_id):
        """ Return the plugin with the specified Id. """

        for plugin in self._plugins:
            if plugin_id == plugin.id:
                if not self._include_plugin(plugin.id):
                    plugin = None

                break

        else:
            plugin = None

        return plugin

    def remove_plugin(self, plugin):
        """ Remove a plugin from the manager. """

        self._plugins.remove(plugin)
        self.plugin_removed = PluginEvent(plugin=plugin)

        return

    def start(self):
        """ Start the plugin manager. """

        for plugin in self._plugins:
            self.start_plugin(plugin)

        return

    def start_plugin(self, plugin=None, plugin_id=None):
        """ Start the specified plugin. """

        plugin = plugin or self.get_plugin(plugin_id)
        if plugin is not None:
            logger.debug('plugin %s starting', plugin.id)
            plugin.activator.start_plugin(plugin)
            logger.debug('plugin %s started', plugin.id)

        else:
            raise SystemError('no such plugin %s' % plugin_id)

        return

    def stop(self):
        """ Stop the plugin manager. """

        # We stop the plugins in the reverse order that they were started.
        stop_order = self._plugins[:]
        stop_order.reverse()

        for plugin in stop_order:
            self.stop_plugin(plugin)

        return

    def stop_plugin(self, plugin=None, plugin_id=None):
        """ Stop the specified plugin. """

        plugin = plugin or self.get_plugin(plugin_id)
        if plugin is not None:
            logger.debug('plugin %s stopping', plugin.id)
            plugin.activator.stop_plugin(plugin)
            logger.debug('plugin %s stopped', plugin.id)

        else:
            raise SystemError('no such plugin %s' % plugin_id)

        return

    #### Protected 'PluginManager' #############################################

    # The plugins that the manager manages!
    _plugins = List(IPlugin)
    def __plugins_changed(self, trait_name, old, new):
        """ Static trait change handler. """

        self._update_plugin_application(old, new)

        return

    def __plugins_items_changed(self, trait_name, old, new):
        """ Static trait change handler. """

        self._update_plugin_application(new.removed, new.added)

        return

    def _include_plugin(self, plugin_id):
        """ Return True if the plugin should be included.

        This is just shorthand for:-

        if self._is_included(plugin_id) and not self._is_excluded(plugin_id):
            ...

        """

        return self._is_included(plugin_id) and not self._is_excluded(plugin_id)

    #### Private protocol ######################################################

    def _is_excluded(self, plugin_id):
        """ Return True if the plugin Id is excluded.

        If no 'exclude' patterns are specified then this method returns False
        for all plugin Ids.

        """

        if len(self.exclude) == 0:
            return False

        for pattern in self.exclude:
            if fnmatch(plugin_id, pattern):
                return True

        return False

    def _is_included(self, plugin_id):
        """ Return True if the plugin Id is included.

        If no 'include' patterns are specified then this method returns True
        for all plugin Ids.

        """

        if len(self.include) == 0:
            return True

        for pattern in self.include:
            if fnmatch(plugin_id, pattern):
                return True

        return False

    def _update_plugin_application(self, removed, added):
        """ Update the 'application' trait of plugins added/removed. """

        for plugin in removed:
            plugin.application = None

        for plugin in added:
            plugin.application = self.application

        return

#### EOF ######################################################################



""" Example code used to test the code browser. """


# Enthought library imports.
from traits.api import HasTraits, Int



class Base(HasTraits):
    """ A class that inherits from 'Hastraits' directly. """

    # Some traits that explicitly specify a trait type.
    x = Int
    y = Int

    def foo(self):
        """ A method. """

    def bar(self):
        """ A method. """

#### EOF ######################################################################



""" A resource protocol for a local file system. """


# Standard library imports.
import errno

# Enthought library imports.
from traits.api import HasTraits, provides

# Local imports.
from .i_resource_protocol import IResourceProtocol
from .no_such_resource_error import NoSuchResourceError


@provides(IResourceProtocol)
class FileResourceProtocol(HasTraits):
    """ A resource protocol for a local file system. """

    ###########################################################################
    # 'IResourceProtocol' interface.
    ###########################################################################

    def file(self, address):
        """ Return a readable file-like object for the specified address. """

        # Opened in binary mode to be consistent with package resources. This
        # means, for example, that line-endings will not be converted.
        try:
            f = open(address, 'rb')

        except IOError as e:
            if e.errno == errno.ENOENT:
                raise NoSuchResourceError(address)

            else:
                raise

        return f

#### EOF ######################################################################



""" A text editor. """


# Standard library imports.
from os.path import basename

# Enthought library imports.
from pyface.workbench.api import TraitsUIEditor
from pyface.api import FileDialog, CANCEL
from traits.api import Code, Instance
from traitsui.api import CodeEditor, Group, Item, View
from traitsui.key_bindings import KeyBinding, KeyBindings
from traitsui.menu import NoButtons

# Local imports.
from .text_editor_handler import TextEditorHandler


def _id_generator():
    """ A generator that returns the next number for untitled files. """

    i = 1
    while True:
        yield(i)
        i += 1

    return

_id_generator = _id_generator()


class TextEditor(TraitsUIEditor):
    """ A text editor. """

    #### 'TextEditor' interface ###############################################

    # The key bindings used by the editor.
    key_bindings = Instance(KeyBindings)

    # The text being edited.
    text = Code

    ###########################################################################
    # 'IEditor' interface.
    ###########################################################################

    def save(self):
        """ Saves the text to disk. """

        # If the file has not yet been saved then prompt for the file name.
        if len(self.obj.path) == 0:
            self.save_as()

        else:
            f = open(self.obj.path, 'w')
            f.write(self.text)
            f.close()

            # We have just saved the file so we ain't dirty no more!
            self.dirty = False

        return

    def save_as(self):
        """ Saves the text to disk after prompting for the file name. """

        dialog = FileDialog(
            parent           = self.window.control,
            action           = 'save as',
            default_filename = self.name,
            wildcard         = FileDialog.WILDCARD_PY
        )
        if dialog.open() != CANCEL:
            # Update the editor.
            self.id   = dialog.path
            self.name = basename(dialog.path)

            # Update the resource.
            self.obj.path = dialog.path

            # Save it!
            self.save()

        return

    ###########################################################################
    # 'TraitsUIEditor' interface.
    ###########################################################################

    def create_ui(self, parent):
        """ Creates the traits UI that represents the editor. """

        ui = self.edit_traits(
            parent=parent, view=self._create_traits_ui_view(), kind='subpanel'
        )

        return ui

    ###########################################################################
    # 'TextEditor' interface.
    ###########################################################################

    def run(self):
        """ Runs the file as Python. """

        # The file must be saved first!
        self.save()

        # Execute the code.
        if len(self.obj.path) > 0:
            view = self.window.get_view_by_id(
                'envisage.plugins.python_shell_view'
            )

            if view is not None:
                view.execute_command(
                    'exec(open(r"%s").read())' % self.obj.path, hidden=False
                )

        return

    def select_line(self, lineno):
        """ Selects the specified line. """

        self.ui.info.text.selected_line = lineno

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    #### Trait initializers ###################################################

    def _key_bindings_default(self):
        """ Trait initializer. """

        key_bindings = KeyBindings(
            KeyBinding(
                binding1    = 'Ctrl-s',
                description = 'Save the file',
                method_name = 'save'
            ),

            KeyBinding(
                binding1    = 'Ctrl-r',
                description = 'Run the file',
                method_name = 'run'
            )
        )

        return key_bindings

    #### Trait change handlers ################################################

    def _obj_changed(self, new):
        """ Static trait change handler. """

        # The path will be the empty string if we are editing a file that has
        # not yet been saved.
        if len(new.path) == 0:
            self.id   = self._get_unique_id()
            self.name = self.id

        else:
            self.id   = new.path
            self.name = basename(new.path)

            f = open(new.path, 'r')
            self.text = f.read()
            f.close()

        return

    def _text_changed(self, trait_name, old, new):
        """ Static trait change handler. """

        if self.traits_inited():
            self.dirty = True

        return

    def _dirty_changed(self, dirty):
        """ Static trait change handler. """

        if len(self.obj.path) > 0:
            if dirty:
                self.name = basename(self.obj.path) + '*'

            else:
                self.name = basename(self.obj.path)

        return

    #### Methods ##############################################################

    def _create_traits_ui_view(self):
        """ Create the traits UI view used by the editor.

        fixme: We create the view dynamically to allow the key bindings to be
        created dynamically (we don't use this just yet, but obviously plugins
        need to be able to contribute new bindings).

        """

        view = View(
            Group(
                Item(
                    'text', editor=CodeEditor(key_bindings=self.key_bindings)
                ),
                show_labels = False
            ),

            id        = 'envisage.editor.text_editor',
            handler   = TextEditorHandler(),
            kind      = 'live',
            resizable = True,
            width     = 1.0,
            height    = 1.0,
            buttons   = NoButtons,
        )

        return view

    def _get_unique_id(self, prefix='Untitled '):
        """ Return a unique id for a new file. """

        id = prefix + str(next(_id_generator))
        while self.window.get_editor_by_id(id) is not None:
            id = prefix + str(next(_id_generator))

        return id

#### EOF ######################################################################



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2006-2007 by Enthought, Inc.
#  All rights reserved.
#
#  Author: Dave Peterson <dpeterson@enthought.com>
#
#-----------------------------------------------------------------------------

"""
A runnable that restores the last opened project.

"""

# Standard library imports
import logging

# Enthought library imports.
from envisage import Runnable
from envisage.workbench.services import IWORKBENCH
from pyface.api import information

# Application imports
from .services import IPROJECT_MODEL, IPROJECT_UI


# Setup a logger for this module.
logger = logging.getLogger(__name__)


class ProjectRunnable(Runnable):
    """
    A runnable that restores the last opened project.

    """

    ##########################################################################
    # 'Runnable' interface.
    ##########################################################################

    #### public interface ####################################################

    def run(self, application):
        """
        Run this runnable.

        Overridden here to: (a) ensure the UI service monitors for the
        closing of the application, and (b) restore the last opened
        project.

        """

        # Ensure our UI service is listening for the application to close.
        # FIXME: This ugly hack (doing this here) is necessary only because
        # this plugin contributes to the workbench plugin and that means the
        # workbench insists on starting us first which means our UI service
        # can't directly reference the workbench service until after
        # everything has been started.
        ui_service = application.get_service(IPROJECT_UI)
        ui_service.listen_for_application_exit()

        # Load the project we were using when we last shutdown.
        model_service = application.get_service(IPROJECT_MODEL)
        location = model_service.preferences.get('project location',
            default=None)
        if location and len(location) > 0:
            logger.info("Opening last project from location [%s]", location)

            try:
                project = model_service.factory.open(location)
            except:
                logger.exception('Error during opening of last project.')
                project = None

            if project is not None:
                model_service.project = project
            else:
                information(self._get_parent_window(application),
                    'Unable to open last project from location:\t\n'
                    '\t%s\n'  % (location) + '\n\n'
                    'The project may no longer exist.',
                    'Can Not Open Last Project',
                    )

        else:
            logger.info('No previous project to open')

        return


    #### protected interface ################################################

    def _get_parent_window(self, application):
        """
        Find and return a reference to the application window.

        If one can not be found, then 'None' is returned.

        """

        window = None
        try:
            workbench = application.get_service(IWORKBENCH)
            window = workbench.active_window.control
        except:
            logger.warn('Unable to retrieve application window reference.')

        return window


#### EOF ######################################################################




""" Classes used to represent classes. """


# Standard library imports.
import ast

# Enthought library imports.
from traits.api import Any, Dict, HasTraits, Instance, Int, List
from traits.api import Str

# Local imports.
from .assign import AssignFactory
from .function import FunctionFactory
from .namespace import Namespace
from ..._compat import STRING_BASE_CLASS


class Klass(Namespace):
    """ A class. """

    #### 'Klass' interface ####################################################

    # The namespace that the class is defined in (this is *usually* a module,
    # but of course classes can be defined anywhere in Python).
    namespace = Instance(Namespace)

    # The line number in the module at which the class appears.
    lineno = Int

    # The name of the class.
    name = Str

    # The class' doc-string (None if there is no doc string, a string if there
    # is).
    doc = Any

    # The class' base classes.
    bases = List(Str)

    # The class' attributes.
    attributes = Dict

    # The class' traits.
    traits = Dict

    # The class' methods.
    methods = Dict

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __str__(self):
        """ Returns an informal string representation of the object. """

        return 'Klass %s at %d' % (self.name, self.lineno)

    def __getstate__(self):
        """ Returns the state of the object for pickling. """

        state = {}
        state['namespace'] = self.namespace
        state['lineno'] = self.lineno
        state['name'] = self.name
        state['doc'] = self.doc
        state['bases'] = self.bases
        state['attributes'] = self.attributes
        state['traits'] = self.traits
        state['methods'] = self.methods

        return state

    def __setstate__(self, state):
        """ Sets the state object duting unpickling. """

        self.namespace = state['namespace']
        self.lineno = state['lineno']
        self.name = state['name']
        self.doc = state['doc']
        self.bases = state['bases']
        self.attributes = state['attributes']
        self.traits = state['traits']
        self.methods = state['methods']

        return


class KlassFactory(HasTraits):
    """ A factory for classes. """

    ###########################################################################
    # 'KlassFactory' interface.
    ###########################################################################

    def from_ast(self, namespace, node):
        """ Creates a class from an AST node. """

        # Create a new class.
        klass = Klass(
            namespace = namespace,
            lineno    = node.lineno,
            name      = node.name,
            doc       = ast.get_docstring(node, clean=False),
            bases     = [self._get_name(base) for base in node.bases]
        )

        # Walk the AST picking out the things we care about!
        KlassVisitor(klass).visit(node)

        return klass

    ###########################################################################
    # Private interface.
    ###########################################################################

    # fixme: This same method is used in 'AssignVisitor'.
    def _get_name(self, node):
        """ Returns the (possibly dotted) name from a node. """

        if isinstance(node, STRING_BASE_CLASS):
            name = node
        elif isinstance(node, ast.Name):
            name = node.id
        else:
            names = [self._get_name(child) for child in ast.walk(node) if child is not node]
            name = '.'.join(names)

        return name


class KlassVisitor(ast.NodeVisitor):
    """ An AST visitor for classes. """

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, klass):
        """ Creates a new visitor. """

        self.klass = klass

        # Factories used to create klasses, functions and assignments from
        # AST nodes.
        self._function_factory = FunctionFactory()
        self._assign_factory   = AssignFactory()

        return

    ###########################################################################
    # 'ASTVisitor' interface.
    ###########################################################################

    def visit_FunctionDef(self, node):
        """ Visits a function node. """

        function = self._function_factory.from_ast(self.klass, node)

        self.klass.locals[node.name] = function
        self.klass.methods[node.name] = function

        return

    def visit_Assign(self, node):
        """ Visits an assignment node. """

        assign = self._assign_factory.from_ast(self.klass, node)

        # Does the assigment look like it *might* be a trait? (i.e., it is NOT
        # an expression or a constant etc.).
        if assign.source:
            assign.is_trait = self.klass.is_trait(assign.source)

        else:
            assign.is_trait = False

        for target in assign.targets:
            self.klass.locals[target] = assign
            self.klass._is_trait[target] = assign.is_trait

            if assign.is_trait:
                self.klass.traits[target] = assign

            else:
                self.klass.attributes[target] = assign

        return

#### EOF ######################################################################



# Standard library imports
import os

# Local imports
from envisage.plugins import remote_editor
from envisage.plugins.remote_editor.communication.client import Client


class EnshellClient(Client):
    """ A Client that is configured to spawn servers using the config file
        for this application.
    """

    server_prefs = (os.path.join(remote_editor.__path__[0],
                    "preferences.ini"),
                    "enthought.remote_editor")



""" The interface for action manager builders. """


# Enthought library imports.
from traits.api import Interface, List

# Local imports.
from .action_set import ActionSet


class IActionManagerBuilder(Interface):
    """ The interface for action manager builders.

    An action manager builder populates action managers (i.e. menus, menu bars
    and tool bars) from the menus, groups and actions defined in its action
    sets.

    """

    # The action sets used by the builder.
    action_sets = List(ActionSet)

    def initialize_action_manager(self, action_manager, root):
        """ Initialize an action manager from the builder's action sets.

        """

#### EOF ######################################################################



""" The *definition* of an action in a tool bar or menu. """


# Enthought library imports.
from traits.api import Str

# Local imports.
from .location import Location


class Action(Location):
    """ The *definition* of an action in a tool bar or menu. """

    #### Action implementation ################################################

    # The action's name (appears on menus and toolbars etc).
    name = Str

    # The name of the class that implements the action.
    class_name = Str

    ###########################################################################
    # 'object' interface
    ###########################################################################

    def __str__(self):
        """ Return the 'informal' string representation of the object. """

        return 'Action(%s)' % self.name

    __repr__ = __str__

#### EOF ######################################################################



""" The 'Orange' plugin """


from envisage.api import Plugin
from traits.api import Bool


class OrangePlugin(Plugin):
    """ The 'Orange' plugin """

    #### 'IPlugin' protocol ####################################################

    # The plugin's unique identifier.
    id = 'orange'

    def start(self):
        """ Start the plugin. """

        self.started = True
        self.stopped = False

        return

    def stop(self):
        """ Stop the plugin. """

        self.started = False
        self.stopped = True

        return

    #### 'BananaPlugin' protocol ###############################################

    started = Bool(False)
    stopped = Bool(False)

#### EOF ######################################################################



# -*- coding: utf-8 -*-
# Standard library imports
import os
import unittest
from time import sleep
from threading import Thread

# ETS imports
from traits.api import Int, Str

# Local imports
from envisage.plugins.remote_editor.communication.client import Client
from envisage.plugins.remote_editor.communication.server import Server
from envisage.plugins.remote_editor.communication.util import \
    get_server_port, LOCK_PATH

from envisage._compat import unicode_str


class TestClient(Client):

    command = Str
    arguments = Str

    error_count = Int(0)

    def handle_command(self, command, arguments):
        self.command = command
        self.arguments = arguments

    def _error_changed(self, error):
        if error:
            self.error_count += 1


class TestThread(Thread):

    def run(self):
        self.server = Server()
        self.server.init()
        self.server.main()


class CommunicationTestCase(unittest.TestCase):

    def setUp(self):
        """ Make sure no old lock files exist prior to run.
        """
        if os.path.exists(LOCK_PATH):
            os.remove(LOCK_PATH)

    def tearDown(self):
        """ Make sure no old lock files are left around.
        """
        if os.path.exists(LOCK_PATH):
            os.remove(LOCK_PATH)

    def testCommunication(self):
        """ Can the Server communicate with Clients and handle errors
            appropriately?
        """
        # Test server set up

        # Does the ping operation work when the Server is not running?
        self.assert_(not Server.ping(get_server_port()))

        # Set up server thread
        serverThread = TestThread()
        serverThread.setDaemon(True)
        serverThread.start()
        sleep(.5)
        self.assert_(os.path.exists(LOCK_PATH))

        # Test normal operation
        tmp = get_server_port()
        self.assert_(Server.ping(tmp))

        client1 = TestClient(self_type='client1', other_type='client2')
        client1.register()
        client2 = TestClient(self_type='client2', other_type='client1')
        client2.register()
        sleep(.5)
        self.assert_(not(client1.orphaned or client2.orphaned))

        client1.send_command("foo", "bar")
        sleep(.1)
        self.assertEqual(client2.command, "foo")
        self.assertEqual(client2.arguments, "bar")

        client1.send_command(unicode_str("Très"), "bien")
        sleep(.1)
        self.assertEqual(client2.command, unicode_str("Très"))
        self.assertEqual(client2.arguments, "bien")


        client1.unregister()
        sleep(.1)
        self.assert_(client1.orphaned and client2.orphaned)

        client1.register()
        sleep(.1)
        self.assert_(not(client1.orphaned or client2.orphaned))

        # Simulated breakage -- does the Server handle unexpected communication
        # failure?

        # Have client1 'die'. We send the dummy command to force its connection
        # loop to terminate after the call to 'stop'. (In Python we can't just
        # kill the thread, which is really what we want to do in this case.)
        client1._communication_thread.stop()
        sleep(.1)
        serverThread.server._send_to(client1._port, "dummy", "")
        sleep(.1)

        # The Server should inform client1 that it could not complete its
        # request
        client2.send_command("foo", "bar")
        sleep(.1)
        self.assert_(client2.orphaned)
        self.assertEqual(client2.error_count, 1)


if __name__ == '__main__':
    """ Run the unittest, but redirect the log to stderr for convenience.
    """
    import logging
    console = logging.StreamHandler()
    console.setLevel(logging.DEBUG)
    logging.getLogger("communication").addHandler(console)
    unittest.main()



""" The interface for protocols that handle resource URLs. """


# Enthought library imports.
from traits.api import Interface


class IResourceProtocol(Interface):
    """ The interface for protocols that handle resource URLs. """

    def file(self, address):
        """ Return a readable file-like object for the specified address.

        Raise a 'NoSuchResourceError' if the resource does not exist.

        e.g.::

          protocol.file('acme.ui.workbench/preferences.ini')

        """

#### EOF ######################################################################



""" Tests for the Egg plugin manager. """


# Enthought library imports.
from envisage.api import EggPluginManager

# Local imports.
from .egg_based_test_case import EggBasedTestCase
from traits.testing.unittest_tools import unittest


class EggPluginManagerTestCase(EggBasedTestCase):
    """ Tests for the Egg plugin manager. """

    ###########################################################################
    # Tests.
    ###########################################################################

    # fixme: Depending how many eggs are on sys.path, this test may take too
    # long to be part of the TDD cycle.
    def test_no_include_or_exclude(self):
        """ no include or exclude """

        # Add all of the eggs in the egg basket.
        self._add_eggs_on_path([self.egg_dir])

        # Make sure that the plugin manager only includes those plugins.
        plugin_manager = EggPluginManager()

        # We don't know how many plugins we will actually get - it depends on
        # what eggs are on sys.path! What we *do* know however is the the 3
        # 'acme' test eggs should be in there!
        ids = [plugin.id for plugin in plugin_manager]

        self.assertTrue('acme.foo' in ids)
        self.assertTrue('acme.bar' in ids)
        self.assertTrue('acme.baz' in ids)

        return

    def test_include_specific(self):
        """ include specific """
        # Add all of the eggs in the egg basket.
        self._add_eggs_on_path([self.egg_dir])

        # The Ids of the plugins that we expect the plugin manager to find.
        expected = ['acme.foo', 'acme.bar']

        # We explicitly limit the plugins to be just the 'acme' test plugins
        # because otherwise the egg plugin manager will pick up *every* plugin
        # in *every* egg on sys.path!
        include = ['acme\.foo', 'acme\.bar']

        # Make sure that the plugin manager only includes those plugins.
        plugin_manager = EggPluginManager(include=include)

        # Make sure the plugin manager found only the required plugins and that
        # it starts and stops them correctly..
        self._test_start_and_stop(plugin_manager, expected)

        return

    def test_include_multiple(self):
        """ include multiple """

        # Add all of the eggs in the egg basket.
        self._add_eggs_on_path([self.egg_dir])

        # The Ids of the plugins that we expect the plugin manager to find.
        expected = ['acme.foo', 'acme.bar', 'acme.baz']

        # We explicitly limit the plugins to be just the 'acme' test plugins
        # because otherwise the egg plugin manager will pick up *every* plugin
        # in *every* egg on sys.path!
        include = ['acme.*']

        # Make sure that the plugin manager only includes those plugins.
        plugin_manager = EggPluginManager(include=include)

        # Make sure the plugin manager found only the required plugins and that
        # it starts and stops them correctly..
        self._test_start_and_stop(plugin_manager, expected)

        return

    def test_exclude_specific(self):
        """ exclude specific """

        # Add all of the eggs in the egg basket.
        self._add_eggs_on_path([self.egg_dir])

        # The Ids of the plugins that we expect the plugin manager to find.
        expected = ['acme.bar']

        # We explicitly limit the plugins to be just the 'acme' test plugins
        # because otherwise the egg plugin manager will pick up *every* plugin
        # in *every* egg on sys.path!
        include = ['acme.*']

        # Now exclude all but 'acme.bar'...
        exclude = ['acme\.foo', 'acme\.baz']

        # Make sure that the plugin manager excludes the specified plugins.
        plugin_manager = EggPluginManager(include=include, exclude=exclude)

        # Make sure the plugin manager found only the required plugins and that
        # it starts and stops them correctly..
        self._test_start_and_stop(plugin_manager, expected)

        return

    def test_exclude_multiple(self):
        """ exclude multiple """

        # Add all of the eggs in the egg basket.
        self._add_eggs_on_path([self.egg_dir])

        # The Ids of the plugins that we expect the plugin manager to find.
        expected = ['acme.foo']

        # We explicitly limit the plugins to be just the 'acme' test plugins
        # because otherwise the egg plugin manager will pick up *every* plugin
        # in *every* egg on sys.path!
        include = ['acme.*']

        # Now exclude every plugin that starts with 'acme.b'.
        exclude = ['acme\.b.*']

        # Make sure that the plugin manager excludes the specified plugins.
        plugin_manager = EggPluginManager(include=include, exclude=exclude)

        # Make sure the plugin manager found only the required plugins and that
        # it starts and stops them correctly..
        self._test_start_and_stop(plugin_manager, expected)

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _test_start_and_stop(self, plugin_manager, expected):
        """ Make sure the plugin manager starts and stops the expected plugins.

        """

        # Make sure the plugin manager found only the required plugins.
        self.assertEqual(expected, [plugin.id for plugin in plugin_manager])

        # Start the plugin manager. This starts all of the plugin manager's
        # plugins.
        plugin_manager.start()

        # Make sure all of the the plugins were started.
        for id in expected:
            plugin = plugin_manager.get_plugin(id)
            self.assertNotEqual(None, plugin)
            self.assertEqual(True, plugin.started)

        # Stop the plugin manager. This stops all of the plugin manager's
        # plugins.
        plugin_manager.stop()

        # Make sure all of the the plugins were stopped.
        for id in expected:
            plugin = plugin_manager.get_plugin(id)
            self.assertNotEqual(None, plugin)
            self.assertEqual(True, plugin.stopped)

        return


# Entry point for stand-alone testing.
if __name__ == '__main__':
    unittest.main()

#### EOF ######################################################################



""" This code has been inspired from the IPython repository

https://github.com/ipython/ipython/blob/2.x/examples/Embedding/internal_ipkernel.py

"""
from ipykernel.connect import connect_qtconsole
from ipykernel.kernelapp import IPKernelApp

from traits.api import Any, HasStrictTraits, Instance, List


def mpl_kernel(gui_backend):
    """ Launch and return an IPython kernel with matplotlib support.

    Parameters
    ----------
    gui_backend -- string or None
      The GUI mode used to initialize the matplotlib mode. For options, see
      the `ipython --matplotlib` help pages. If None, the kernel is initialized
      without GUI support.
    """

    kernel = IPKernelApp.instance()

    argv = ['python']
    if gui_backend is not None:
        argv.append('--matplotlib={}'.format(gui_backend))
    kernel.initialize(argv)

    return kernel


class InternalIPKernel(HasStrictTraits):
    """ Represents an IPython kernel and the consoles attached to it.
    """

    #: The IPython kernel.
    ipkernel = Instance(IPKernelApp)

    #: A list of connected Qt consoles.
    consoles = List()

    #: The kernel namespace.
    #: Use `Any` instead of `Dict` because this is an IPython dictionary
    #: object.
    namespace = Any()

    #: The values used to initialize the kernel namespace.
    #: This is a list of tuples (name, value).
    initial_namespace = List()

    def init_ipkernel(self, gui_backend):
        """ Initialize the IPython kernel.

        Parameters
        ----------
        gui_backend -- string
          The GUI mode used to initialize the matplotlib mode. For options, see
          the `ipython --matplotlib` help pages.
        """
        # Start IPython kernel with GUI event loop and mpl support
        self.ipkernel = mpl_kernel(gui_backend)

        # This application will also act on the shell user namespace
        self.namespace = self.ipkernel.shell.user_ns
        self.namespace.update(dict(self.initial_namespace))

    def new_qt_console(self):
        """ Start a new qtconsole connected to our kernel. """
        console = connect_qtconsole(self.ipkernel.connection_file,
                                    argv=['--no-confirm-exit'])
        self.consoles.append(console)
        return console

    def cleanup_consoles(self):
        """ Kill all existing consoles. """
        for c in self.consoles:
            c.kill()
        self.consoles = []

    def shutdown(self):
        """ Shutdown the kernel.

        Existing IPython consoles are killed first.
        """
        if self.ipkernel is not None:
            self.cleanup_consoles()
            self.ipkernel.shell.exit_now = True
            self.ipkernel.cleanup_connection_file()
            self.ipkernel = None



from .preferences_category import PreferencesCategory
from .preferences_pane import PreferencesPane
from .task_extension import TaskExtension
from .task_factory import TaskFactory
from .task_window import TaskWindow
from .tasks_application import TasksApplication



""" A plugin for controlling a remote editor. """

# Enthought library imports.
from envisage.api import Plugin
from envisage.plugins.remote_editor.api import IRemoteEditor
from traits.api import List, Instance, Any, on_trait_change

# Local imports
from .envisage_remote_editor import EnvisageRemoteEditorController \
    as RemoteEditorController

ID = 'envisage.plugins.remote_editor'


class RemoteEditorPlugin(Plugin):
    """ A plugin for controlling a remote editor. """

    # Extension point Ids.
    REMOTE_EDITOR     = ID
    ACTION_SETS       = 'envisage.ui.workbench.action_sets'
    PREFERENCES       = 'envisage.preferences'
    PREFERENCES_PAGES = 'envisage.ui.workbench.preferences_pages'

    # Our remote controller for the editor
    remote_controller = Instance(RemoteEditorController)

    # The shell and editor commands
    server_prefs = Any

    #### 'IPlugin' interface ##################################################

    # The plugin's unique identifier.
    id = 'envisage.plugins.remote_editor'

    # The plugin's name (suitable for displaying to the user).
    name = 'Remote editor'

    #### Extension points offered by this plugin ##############################

    #### Contributions to extension points made by this plugin ################

    # Our action sets.
    action_sets = List(contributes_to=ACTION_SETS)

    # Preferences pages.
    # FIXME: Create a UI for remote editor preferences
    #preferences_pages = List(contributes_to=PREFERENCES_PAGES)

    # Preferences.
    preferences = List(contributes_to=PREFERENCES)

    def _action_sets_default(self):
        """ Trait initializer. """
        from envisage.plugins.remote_editor.actions import \
            RemoteEditorActionSet
        return [ RemoteEditorActionSet ]

    def _preferences_default(self):
        """ Trait initializer. """
        return [ 'pkgfile://%s/preferences.ini' % ID ]

    def _preferences_pages_default(self):
        """ Trait initializer. """
        from envisage.plugins.remote_editor.preference_pages \
            import RemoteEditorPreferencesPage
        return [ RemoteEditorPreferencesPage ]

    ###########################################################################
    # Private interface.
    ###########################################################################

    # Create the central server for spawning shells and editors.
    @on_trait_change('application:started')
    def _create_server(self):
        """ Create the central server for spawning shells and editors and
            register the controller as an envisage service.
        """
        # Register our client to the server. If the server does not exist, this
        # will create it.
        self.remote_controller = RemoteEditorController(
            application=self.application)

        # XXX I don't like this at all
        if self.server_prefs:
            self.remote_controller.server_prefs = self.server_prefs

        self.remote_controller.register()

        self.application.register_service(IRemoteEditor, self.remote_controller)

    @on_trait_change('application:stopping')
    def _unregister_from_server(self):
        """ Unregister this client from the server.
        """
        self.remote_controller.unregister()

#### EOF ######################################################################



""" The Envisage workbench plugin. """


# Enthought library imports.
from envisage.api import ExtensionPoint, Plugin, ServiceOffer
from traits.api import Callable, List


# This module's package.
PKG = '.'.join(__name__.split('.')[:-1])


class WorkbenchPlugin(Plugin):
    """ The Envisage workbench plugin.

    The workbench plugin uses the PyFace workbench to provide the basis of an
    IDE-like user interface. The interface is made up of perspectives, views
    and editors.

    Note that this is not intended to be a 'general-purpose' plugin for user
    interfaces - it provides an IDE-like style and that is all. If your
    application requires another style of interface then write another plugin
    (you can still re-use all the menu, group and action contribution stuff!).

    """

    # The Ids of the extension points that this plugin offers.
    ACTION_SETS              = PKG + '.action_sets'
    PERSPECTIVES             = PKG + '.perspectives'
    PREFERENCES_PAGES        = PKG + '.preferences_pages'
    WORKBENCH_SERVICE_OFFERS = PKG + '.service_offers'
    VIEWS                    = PKG + '.views'

    # The Ids of the extension points that this plugin contributes to.
    PREFERENCES    = 'envisage.preferences'
    SERVICE_OFFERS = 'envisage.service_offers'

    #### 'IPlugin' interface ##################################################

    # The plugin's unique identifier.
    id = 'envisage.ui.workbench'

    # The plugin's name (suitable for displaying to the user).
    name = 'Workbench'

    #### Extension points offered by this plugin ##############################

    action_sets = ExtensionPoint(
        List(Callable), id=ACTION_SETS, desc="""

        An action set contains the toobars, menus, groups and actions that you
        would like to add to top-level workbench windows (i.e. the main
        application window). You can create new toolbars, menus and groups
        and/or add to existing ones.

        Each contribution to this extension point must be a factory that
        creates an action set, where 'factory' means any callable with the
        following signature::

          callable(**traits) -> IActionSet

        The easiest way to contribute such a factory is to create a class
        that derives from 'envisage.ui.action.api.ActionSet'.

        """
    )

    perspectives = ExtensionPoint(
        List(Callable), id=PERSPECTIVES, desc="""

        A perspective is simply an arrangment of views around the (optionally
        hidden) editor area.

        Each contribution to this extension point must be a factory that
        creates a perspective, where 'factory' means any callable with the
        following signature::

          callable(**traits) -> IPerspective

        The easiest way to contribute such a factory is to create a class
        that derives from 'pyface.workbench.api.IPerspective'.

        """
    )

    preferences_pages = ExtensionPoint(
        List(Callable), id=PREFERENCES_PAGES, desc="""

        A preferences page appears in the preferences dialog to allow the user
        to manipulate some preference values.

        Each contribution to this extension point must be a factory that
        creates a preferences page, where 'factory' means any callable with the
        following signature::

          callable(**traits) -> IPreferencesPage

        The easiest way to contribute such a factory is to create a class
        that derives from 'apptools.preferences.ui.api.IPreferencesPage'.

        """
    )

    service_offers = ExtensionPoint(
        List(ServiceOffer),
        id   = WORKBENCH_SERVICE_OFFERS,
        desc = """

        Services are simply objects that a plugin wants to make available to
        other plugins. This extension point allows you to offer 'per
        window' services that are created 'on-demand' (where 'on demand' means
        the first time somebody looks up a service of the appropriate
        protocol).
        .

        e.g.

        my_service_offer = ServiceOffer(
            protocol   = 'acme.IMyService',
            factory    = an_object_or_a_callable_that_creates_one,
            properties = {'a dictionary' : 'that is passed to the factory'}
        )

        Any properties specified are passed as keywrod arguments to the
        factory, i.e. the factory signature is::

          callable(**properties)

        """
    )

    views = ExtensionPoint(
        List(Callable), id=VIEWS, desc="""

        A view provides information to the user to support their current
        task. Views can contain anything you like(!) and are arranged around
        the (optionally hidden) editor area. The user can re-arrange views as
        he/she sees fit.

        Each contribution to this extension point must be a factory that
        creates a view, where 'factory' means any callable with the following
        signature::

          callable(**traits) -> IView

        The easiest way to contribute such a factory is to create a class
        that derives from 'pyface.workbench.api.View'.

        It is also common to use a simple function (especially when a view
        is a representation of a service) e.g::

            def foo_view_factory(**traits):
                ' Create a view that is a representation of a service. '
                foo = self.application.get_service('IFoo')

                return FooView(foo=foo, **traits)

        """
    )

    #### Contributions to extension points made by this plugin ################

    my_action_sets = List(contributes_to=ACTION_SETS)

    def _my_action_sets_default(self):
        """ Trait initializer. """

        from .default_action_set import DefaultActionSet

        return [DefaultActionSet]

    my_preferences = List(contributes_to=PREFERENCES)

    def _my_preferences_default(self):
        """ Trait initializer. """

        return ['pkgfile://envisage.ui.workbench/preferences.ini']

    my_preferences_pages = List(contributes_to=PREFERENCES_PAGES)

    def _my_preferences_pages_default(self):
        """ Trait initializer. """

        from .workbench_preferences_page import WorkbenchPreferencesPage

        return [WorkbenchPreferencesPage]

    my_service_offers = List(contributes_to=SERVICE_OFFERS)

    def _my_service_offers_default(self):
        """ Trait initializer. """

        preferences_manager_service_offer = ServiceOffer(
            protocol = 'apptools.preferences.ui.preferences_manager'
                       '.PreferencesManager',
            factory  = self._create_preferences_manager_service
        )

        workbench_service_offer = ServiceOffer(
            protocol = 'envisage.ui.workbench.workbench.Workbench',
            factory  = self._create_workbench_service
        )

        return [preferences_manager_service_offer, workbench_service_offer]

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _create_preferences_manager_service(self, **properties):
        """ Factory method for the preferences manager service. """

        from apptools.preferences.ui.api import PreferencesManager

        preferences_manager = PreferencesManager(
            pages=[factory() for factory in self.preferences_pages]
        )

        return preferences_manager

    def _create_workbench_service(self, **properties):
        """ Factory method for the workbench service. """

        # We don't actually create the workbench here, we just return a
        # reference to it.
        #
        # fixme: This guard is really just for testing when we have the
        # workbench plugin as a source egg (i.e. if the egg is on our path
        # then we get the plugin for any egg-based application, even if it is
        # not a workbench application!).
        return getattr(self.application, 'workbench', None)

### EOF ######################################################################



""" A test class used to test categories. """


# Enthought library imports.
from traits.api import HasTraits, Int


class BarCategory(HasTraits):
    y = Int

#### EOF ######################################################################



import unittest

try:
    import IPython  # noqa
except ImportError:
    from nose.plugins.skip import SkipTest
    raise SkipTest('IPython not available')

from ipykernel.kernelapp import IPKernelApp
from envisage.plugins.ipython_kernel.internal_ipkernel import InternalIPKernel


class TestInternalIPKernel(unittest.TestCase):

    def tearDown(self):
        IPKernelApp.clear_instance()

    def test_lifecycle(self):
        kernel = InternalIPKernel()
        self.assertIsNone(kernel.ipkernel)

        kernel.init_ipkernel(gui_backend=None)
        self.assertIsNotNone(kernel.ipkernel)
        self.assertIsInstance(kernel.ipkernel, IPKernelApp)

        kernel.new_qt_console()
        kernel.new_qt_console()
        self.assertEqual(len(kernel.consoles), 2)

        kernel.shutdown()
        self.assertIsNone(kernel.ipkernel)
        self.assertEqual(len(kernel.consoles), 0)

    def test_initial_namespace(self):
        kernel = InternalIPKernel(initial_namespace=[('x', 42.1)])
        kernel.init_ipkernel(gui_backend=None)
        self.assertIn('x', kernel.namespace)
        self.assertEqual(kernel.namespace['x'], 42.1)
        kernel.shutdown()



""" The developer UI plugin. """


# Enthought library imports.
from envisage.api import Plugin
from traits.api import List


class DeveloperUIPlugin(Plugin):
    """ The developer UI plugin.

    This plugin contains the UI part of the tools that (hopefully) help
    developers to inspect and debug a running Envisage workbench application.

    """

    # The plugin Id.
    ID = 'envisage.developer.ui'

    # Extension points Ids.
    PERSPECTIVES = 'envisage.ui.workbench.perspectives'
    VIEWS        = 'envisage.ui.workbench.views'

    #### 'IPlugin' interface ##################################################

    # The plugin's unique identifier.
    id   = ID

    # The plugin's name (suitable for displaying to the user).
    name = 'Developer UI'

    #### 'DeveloperUIPlugin' interface ########################################

    #### Extension points offered by this plugin ##############################

    # None.

    #### Contributions to extension points made by this plugin ################

    perspectives = List(contributes_to=PERSPECTIVES)

    def _perspectives_default(self):
        """ Trait initializer. """

        from envisage.developer.ui.perspective.api import (
            DeveloperPerspective
        )

        return [DeveloperPerspective]

    views = List(contributes_to=VIEWS)

    def _views_default(self):
        """ Trait initializer. """

        from .view.api import (
            ApplicationBrowserView,
            ExtensionRegistryBrowserView,
            ServiceRegistryBrowserView
        )

        views = [
            ApplicationBrowserView,
            ExtensionRegistryBrowserView,
            ServiceRegistryBrowserView
        ]

        return views

#### EOF ######################################################################



# Standard library imports
from errno import EINTR
import os
import select
import socket
from subprocess import Popen
import sys

import csv
from io import StringIO

# ETS imports
from traits.etsconfig.api import ETSConfig

# Local imports
from envisage._compat import STRING_BASE_CLASS


# An obscure ASCII character that we used as separators in socket streams
MESSAGE_SEP = chr(7) # 'bell' character

# The location of the server lock file and the communication log
LOCK_PATH = os.path.join(ETSConfig.application_data,
                         'remote_editor_server.lock')
LOG_PATH = os.path.join(ETSConfig.application_data, 'remote_editor_server.log')

def encode(s):
    return s.encode('utf-8')
def decode(s):
    return s.decode('utf-8')


def quoted_split(s):
    f = StringIO(s)
    split = next(csv.reader(f, delimiter=' ', quotechar='"'))
    return split


def spawn_independent(command, shell=False):
    """ Given a command suitable for 'Popen', open the process such that if this
        process is killed, the spawned process survives.

        `command` is either a list of strings, with the first item in the list being
        the executable and the rest being its arguments, or a single string containing
        the executable and its arguments.  In the latter case, any argument that
        contains spaces must be delimited with double-quotes.
    """
    if sys.platform == 'win32':
        if isinstance(command, STRING_BASE_CLASS):
            command = 'start /b ' + command
        else:
            command.insert(0, 'start')
            command.insert(1, '/b')
        Popen(command, shell=True)
    elif sys.platform == 'darwin':
        pid = os.fork()
        if pid:
            return
        else:
            os.setpgrp()
            if isinstance(command, STRING_BASE_CLASS):
                tmp = quoted_split(command)
            else:
                tmp = command
            os.execv(tmp[0], tmp)
    else:
        pid = os.fork()
        if pid:
            return
        else:
            os.setpgrp()
            Popen(command, shell=shell)
            sys.exit(0)

def get_server_port():
    """ Reads the server port from the lock file. If the file does not exist
        returns -1.
    """
    if os.path.exists(LOCK_PATH):
        with open(LOCK_PATH, 'rt') as f:
            return int(f.read().strip())
    else:
        return -1


def accept_no_intr(sock):
    """ Call sock.accept such that if it is interrupted by an EINTR
        ("Interrupted system call") signal or a KeyboardInterrupt exception, it
        is re-called.
    """
    while True:
        try:
            return sock.accept()
        except socket.error as err:
            if err.args[0] != EINTR:
                raise
        except KeyboardInterrupt:
            pass


def send(sock, command, arguments=""):
    """ Send a command with arguments (both strings) through a socket. This
        information is encoded with length information to ensure that everything
        is received.
    """
    sent, total = 0, 0
    msg = command + MESSAGE_SEP + arguments
    msg = str(len(msg)) + MESSAGE_SEP + msg
    length = len(msg)

    msg = encode(msg)
    while total < length:
        msg = msg[sent:]
        sent = sock.send(msg)
        if not sent:
            raise socket.error
        total += sent


def send_port(port, command, arguments="", timeout=None):
    """ A send a command to a port. Convenience function that uses 'send'.
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    if timeout is not None:
        sock.settimeout(timeout)
    errno = sock.connect_ex(('localhost', port))
    if errno:
        sock.close()
        return False

    try:
        send(sock, command, arguments)
    except socket.error:
        return False
    finally:
        try:
            sock.shutdown(socket.SHUT_WR)
        except:
            pass
    return True


def receive(sock):
    """ Receive a command with arguments from a socket that was previously sent
        information with 'send'.
    """
    index = -1
    chunk, length = encode(""), encode("")
    received = False

    while not received:
        # Use select to query the socket for activity before calling sock.recv.
        # Choose a timeout of 60 seconds for now.
        readers, writers, in_error = select.select([sock], [], [], 60)

        for rsock in readers:
            while index == -1:
                length += chunk
                chunk = rsock.recv(4096)
                if not chunk:
                    raise socket.error
                index = chunk.find(encode(MESSAGE_SEP))
            length = int(length + chunk[:index])

            msg = chunk[index+1:]
            while len(msg) < length:
                chunk = rsock.recv(length - len(msg))
                if not chunk:
                    raise socket.error
                msg += chunk
            received = True

    command, sep, arguments = msg.partition(encode(MESSAGE_SEP))
    return decode(command), decode(arguments)
